import json
from pathlib import Path
import subprocess
from typing import List, Optional
import threading
import concurrent.futures

import typer
from tqdm import tqdm
from tabulate import tabulate
from . import db, utils

cli = typer.Typer()
lock = threading.Lock()

# create tables
db.init()


@cli.command()
def index(
    url: str,
    lang: str = "es",
    verbose: bool = False,
    max_threads: Optional[int] = None,
    cookies_from: Optional[str] = None,
    proxy: Optional[str] = None,
):
    print(f"Fetching video urls from {url}...")
    video_ids = list(
        utils.get_video_ids(
            url, cookies_from=cookies_from, lang=lang, verbose=verbose, proxy=proxy
        )
    )
    new_ids = list(db.filter_existing_video_ids(video_ids, lang))

    # download and index subs in parallel
    with concurrent.futures.ThreadPoolExecutor(max_threads) as executor:
        tasks = (
            executor.submit(
                utils.download_and_process_subtitles,
                video_id,
                lang,
                lock=lock,
                verbose=verbose,
                proxy=proxy,
            )
            for video_id in new_ids
        )
        try:
            for future in tqdm(
                concurrent.futures.as_completed(tasks),
                desc="Downloading subtitles",
                total=len(new_ids),
            ):
                try:
                    future.result()
                except Exception as e:
                    typer.echo(f"Error downloading subtitles: {e}")
        except KeyboardInterrupt:
            typer.echo("Aborting...")
            executor.shutdown(cancel_futures=True)
            raise typer.Exit(1)


@cli.command()
def search(
    text: str,
    user: str = None,
    lang: str = None,
    format: str = "",
    download_parts: bool = False,
    spacing_secs: int = 5,
    proxy: Optional[str] = None,
    merge: bool = False,
):
    results = []
    for row in db.search(text, uploader_id=user, lang=lang):
        results.append(row)

    if not results:
        print("No results")
        return

    if download_parts:
        print(f"Downloading {len(results)} parts...")
        for row in results:
            utils.download_part(
                row,
                folder=f"output-{text.replace(' ', '-').lower()}",
                spacing_secs=spacing_secs,
                proxy=proxy,
            )
        return

    if format == "json":
        print(json.dumps(results, indent=2))
    else:
        print(tabulate(results, headers="keys", maxcolwidths=20))


@cli.command()
def list_channels(format: str = ""):
    channels = list(db.get_channels())

    if format == "json":
        print(json.dumps(channels, indent=2))
    else:
        print(tabulate(channels, headers="keys", tablefmt="simple"))


@cli.command()
def stats(format: str = ""):
    _db = db.get_db()

    stats = {
        "channels": _db["channels"].count,
        "videos": _db["videos"].count,
    }

    if format == "json":
        print(json.dumps(stats, indent=2))
    else:
        print(tabulate(stats.items()))


@cli.command()
def remove_channel(
    uploader_id: str = typer.Argument(help="Channel user handle (ie. @user)"),
):
    db.delete_channel(uploader_id)


@cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    help="Run datasette, supports all datasette options",
)
def server(ctx: typer.Context, port: str = "8001"):
    try:
        import datasette
    except ImportError:
        typer.echo("datasette not installed. Run `pip install yt-supercut[datasette]`")
        raise typer.Exit(1)

    args = [
        "datasette",
        db.DB_PATH,
        "--metadata",
        str(Path().parent / "metadata.json"),
        "--port",
        port,
        *ctx.args,
    ]

    print(
        f"When server is ready, open http://localhost:{port}/youtube/subtitles_with_videos"
    )
    subprocess.run(args)


@cli.command()
def build_transcripts(
    output: str = ".",
):
    for video in db.get_video_languages():
        video_id = video["video_id"]
        lang = video["lang"]

        print(f"Building transcript for {video_id} in {lang}...")
        utils.build_transcript(video_id, lang=lang, path=output)


if __name__ == "__main__":
    cli()



import os
import re
import math
import json
import tempfile
from pathlib import Path

import webvtt
import appdirs
import frontmatter
from yt_dlp import YoutubeDL
from yt_dlp.utils import download_range_func, ExistingVideoReached
from . import db


CONFIG_DIR = appdirs.user_config_dir("yt_supercut")
Path(CONFIG_DIR).mkdir(parents=True, exist_ok=True)


def get_video_ids(url, cookies_from=None, lang=None, verbose=True, proxy=None):
    stripped_url = re.sub(r"[^\w]", "_", url)
    archive_path = Path(CONFIG_DIR) / f"{stripped_url}.{lang}.txt"

    opts = {
        "skipdownload": True,
        "no_warnings": True,
        "noprogress": True,
        "dumpjson": True,
        "quiet": not verbose,
        "extract_flat": "in_playlist",
        "extractor_args": {
            "youtube": {"skip": ["dash", "hls"]},
            "youtubetab": {"approximate_date": ["timestamp"]},
        },
        "break_on_existing": True,
        "break_per_url": True,
        "force_write_download_archive": True,
        "download_archive": str(archive_path),
    }

    if cookies_from:
        opts.update(
            {
                "cookiesfrombrowser": cookies_from.split(","),
            }
        )

    if proxy:
        opts.update(
            {
                "proxy": proxy,
            }
        )

    with YoutubeDL(opts) as ydl:
        try:
            ydl.extract_info(url, download=False)
        except ExistingVideoReached:
            pass

    with archive_path.open() as f:
        for line in f.readlines():
            _, video_id = line.strip().split(" ")
            yield video_id


def download_and_process_subtitles(
    video_id, lang, cookies_from=None, lock=None, verbose=False, proxy=None
):
    with tempfile.TemporaryDirectory() as tmpdir:
        opts = {
            "skip_download": True,
            "no_warnings": True,
            "writeautomaticsub": True,
            "writeinfojson": True,
            "convertsubtitles": True,
            "noprogress": True,
            "subtitlesformat": "vtt",
            "subtitleslangs": [lang, "-live_chat"],
            "quiet": not verbose,
            "outtmpl": os.path.join(tmpdir, "%(id)s.%(ext)s"),
        }

        if cookies_from:
            opts.update(
                {
                    "cookiesfrombrowser": cookies_from.split(","),
                }
            )

        if proxy:
            opts.update(
                {
                    "proxy": proxy,
                }
            )

        with YoutubeDL(opts) as yt:
            # download subtitles
            yt.download([f"https://www.youtube.com/watch?v={video_id}"])

        with lock:
            # process subtitles
            # inside a lock, since sqlite3 writes are not thread-safe
            process_subtitles(video_id, lang, tmpdir)


def process_subtitles(video_id, lang, folder):
    filename_sub = os.path.join(folder, f"{video_id}.{lang}.vtt")
    filename_infojson = os.path.join(folder, f"{video_id}.info.json")

    if not os.path.exists(filename_sub):
        db.add_video_language(video_id, lang, available=False)
        return

    with open(filename_infojson) as f:
        info = json.load(f)

    subtitles = []
    filepath = os.path.join(folder, f"{video_id}.{lang}.vtt")

    for caption in webvtt.read(filepath):
        if caption.end_in_seconds - caption.start_in_seconds < 0.5:
            continue

        subtitles.append(
            {
                "video_id": video_id,
                "lang": lang,
                "start_seconds": math.floor(caption.start_in_seconds),
                "end_seconds": math.ceil(caption.end_in_seconds),
                "start_time": caption.start,
                "end_time": caption.end,
                "text": caption.text.split("\n")[0],
            }
        )

    db.clear_subtitles(video_id, lang)
    db.add_subtitles(video_id, lang, subtitles)
    db.add_video_info(info)
    db.add_channel_info(info)


def download_part(info, folder="output", spacing_secs=5, proxy=None):
    print(f"Dowloading {info}")

    if not os.path.exists(folder):
        os.mkdir(folder)

    opts = {
        "ignoreerrors": True,
        "force_keyframes_at_cuts": True,
        "download_ranges": download_range_func(
            None,
            [
                (
                    info["start_seconds"] - spacing_secs,
                    info["end_seconds"] + spacing_secs,
                )
            ],
        ),
        "download_archive": os.path.join(folder, "archive.txt"),
        "outtmpl": os.path.join(
            folder, "%(title)s.%(id)s.%(start_time)s-%(end_time)s.%(ext)s"
        ),
    }

    if proxy:
        opts.update(
            {
                "proxy": proxy,
            }
        )

    with YoutubeDL(opts) as yt:
        yt.download([info["link"]])


def build_transcript(video_id, lang=None, path: str = "."):
    video = db.get_video(video_id)
    fn = os.path.join(path, f"{video_id}.{lang}.md")

    if os.path.exists(fn):
        print(f"Transcript for {video_id} in {lang} already exists.")
        return

    if not video:
        print(f"Video {video_id} not found in the database.")
        return

    subtitles = db.get_subtitles(video_id, lang)
    if not subtitles:
        print(f"No subtitles found for {video_id} in {lang}")
        return

    transcript = ""
    for idx, subtitle in enumerate(subtitles):
        if idx % 25 == 0:
            transcript += (
                f"\n[{subtitle['start_time']} / ?t={subtitle['start_seconds']}]\n"
            )
        transcript += f"{subtitle['text']} "

    with open(fn, "wb") as f:
        p = frontmatter.Post(
            transcript,
            **{
                "title": video["video_title"],
                "lang": lang,
                "url": video["video_url"],
            },
        )
        frontmatter.dump(p, f)



import os
from datetime import datetime
from sqlite_utils import Database

DB_PATH = os.getenv("DB_PATH", "youtube.db")


def get_db() -> Database:
    return Database(DB_PATH)


def init():
    db = get_db()

    db["channels"].create(
        {
            "uploader_id": str,
            "channel_name": str,
            "channel_url": str,
        },
        pk="uploader_id",
        not_null={"channel_url"},
        if_not_exists=True,
    )

    db["videos"].create(
        {
            "video_id": str,
            "video_title": str,
            "video_url": str,
            "uploader_id": str,
            "upload_date": str,
        },
        pk="video_id",
        not_null={"video_title", "video_url"},
        if_not_exists=True,
        foreign_keys=[
            ("uploader_id", "channels"),
        ],
    )
    db["videos"].create_index(["uploader_id"], if_not_exists=True)

    db["subtitles"].create(
        {
            "subtitle_id": int,
            "video_id": str,
            "start_time": str,
            "end_time": str,
            "start_seconds": int,
            "end_seconds": int,
            "lang": str,
            "text": str,
        },
        pk="subtitle_id",
        not_null={
            "start_seconds",
            "end_seconds",
            "start_time",
            "end_time",
            "lang",
            "text",
        },
        if_not_exists=True,
        foreign_keys=[("video_id", "videos")],
    ).enable_fts(["text"], create_triggers=True, replace=True)

    db["video_languages"].create(
        {
            "video_id": str,
            "lang": str,
            "available": bool,
        },
        pk=("video_id", "lang"),
        foreign_keys=[
            ("video_id", "videos"),
        ],
        defaults={"available": True},
        not_null={"lang"},
        if_not_exists=True,
    )

    db.create_view(
        "subtitles_with_videos",
        """
        select
            s.subtitle_id,
            v.video_id,
            v.uploader_id,
            v.video_title,
            v.upload_date,
            c.channel_name,
            s.start_seconds,
            s.end_seconds,
            s.lang,
            s.text,
            v.video_url || "&start=" || (s.start_seconds-4) || "&end=" || (s.end_seconds+2) as link
        from subtitles s
        join videos v on s.video_id = v.video_id
        join channels c ON v.uploader_id = c.uploader_id
    """,
        replace=True,
    )


def add_channel_info(info):
    db = get_db()
    db["channels"].insert(
        {
            "uploader_id": info["uploader_id"],
            "channel_name": info["uploader"],
            "channel_url": info["channel_url"],
        },
        replace=True,
    )


def add_video_info(info):
    db = get_db()
    db["videos"].insert(
        {
            "video_id": info["id"],
            "video_title": info["title"],
            "video_url": info["webpage_url"],
            "uploader_id": info["uploader_id"],
            "upload_date": datetime.strptime(info["upload_date"], "%Y%m%d"),
        },
        replace=True,
    )


def add_subtitles(video_id, lang, items):
    db = get_db()
    db["subtitles"].insert_all(items)
    db["video_languages"].insert(
        {
            "video_id": video_id,
            "lang": lang,
            "available": True,
        },
        replace=True,
    )


def clear_subtitles(video_id, lang):
    db = get_db()
    db["subtitles"].delete_where(
        "video_id = :video_id AND lang = :lang",
        {
            "video_id": video_id,
            "lang": lang,
        },
    )
    db["video_languages"].delete_where(
        "video_id = :video_id AND lang = :lang",
        {
            "video_id": video_id,
            "lang": lang,
        },
    )
    db.conn.commit()


def add_video_language(video_id, lang, available=True):
    db = get_db()
    db["video_languages"].insert(
        {
            "video_id": video_id,
            "lang": lang,
            "available": available,
        },
        replace=True,
    )


def search(text, uploader_id=None, lang=None):
    db = get_db()
    where = []
    where_args = {}

    where.append(
        f"subtitle_id in (select rowid from subtitles_fts where subtitles_fts match :text)"
    )
    where_args["text"] = text

    if lang:
        where.append(f"lang = :lang")
        where_args["lang"] = lang

    if uploader_id:
        where.append(
            f"video_id in (select video_id from videos where uploader_id = :uploader_id)"
        )
        where_args["uploader_id"] = uploader_id

    for row in db["subtitles_with_videos"].rows_where(
        " AND ".join(where),
        where_args,
        order_by="video_id, start_seconds ASC",
    ):
        yield row


def get_channels():
    db = get_db()
    for channel in db["channels"].rows:
        yield channel


def get_channel(uploader_id):
    db = get_db()
    for row in db["channels"].rows_where("uploader_id = ?", [uploader_id]):
        return row


def get_videos():
    db = get_db()
    for video in db["videos"].rows:
        yield video


def get_video_languages():
    db = get_db()
    for video_language in db["video_languages"].rows:
        yield video_language


def filter_existing_video_ids(video_ids, lang):
    db = get_db()
    db.execute(
        "CREATE TEMPORARY TABLE IF NOT EXISTS tmp (video_id TEXT NOT NULL PRIMARY KEY)"
    )
    db["tmp"].insert_all([{"video_id": video_id} for video_id in video_ids])
    for row in db["tmp"].rows_where(
        "video_id not in (select video_id from video_languages where lang = :lang)",
        {"lang": lang},
    ):
        yield row["video_id"]
    db["tmp"].drop()


def get_video(video_id):
    db = get_db()
    for row in db["videos"].rows_where("video_id = ?", [video_id]):
        return row


def delete_video(video_id):
    db = get_db()
    db["subtitles"].delete_where("video_id = ?", [video_id])
    db["video_languages"].delete_where("video_id = ?", [video_id])
    db["videos"].delete(video_id)


def delete_channel(uploader_id):
    db = get_db()
    db["videos"].delete_where("uploader_id = ?", [uploader_id])
    db["subtitles"].delete_where(
        "video_id in (select video_id from videos where uploader_id = ?)", [uploader_id]
    )
    db["video_languages"].delete_where(
        "video_id in (select video_id from videos where uploader_id = ?)", [uploader_id]
    )
    db["channels"].delete(uploader_id)


def get_subtitles(video_id, lang=None):
    db = get_db()
    where = "video_id = :video_id"
    where_args = {"video_id": video_id}

    if lang:
        where += " AND lang = :lang"
        where_args["lang"] = lang

    for row in db["subtitles"].rows_where(
        where, where_args, order_by="start_seconds ASC"
    ):
        yield row
